# -*- coding: utf-8 -*-
import curses
import pdb
import sys


screen = None


def init(stdscr):
    global screen
    screen = stdscr


def finalize(stdscr=None):
    if not stdscr and not screen:
        raise Exception('either call init() first or provide a window object')
    stdscr = screen if screen and not stdscr else stdscr
    curses.nocbreak()
    stdscr.keypad(0)
    curses.echo()
    curses.endwin()


def debug(stdscr=None):
    if not stdscr and not screen:
        raise Exception('either call init() first or provide a window object')
    stdscr = screen if screen and not stdscr else stdscr
    finalize(stdscr)

    debugger = pdb.Pdb()
    debugger.reset()
    debugger.do_where(None)
    users_frame = sys._getframe().f_back  # One frame up, outside this function
    debugger.interaction(users_frame, None)


# Use with:
# import cursutils
# cursutils.init(stdscr)   # where stdscr is a `curses` Window object
# cursutils.debug()



# -*- coding: utf-8 -*-


class ArgumentException(Exception):
    pass


class NotAGitRepositoryException(Exception):
    pass


class GitNotFoundException(Exception):
    pass


class CommandErrorException(Exception):
    pass



# -*- coding: utf-8 -*-


class Event(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.iteritems():
            setattr(self, k, v)


class Trigger(object):
    def __init__(self):
        self.callbacks = set()

    def subscribe(self, callback):
        self.callbacks.add(callback)

    def unsubscribe(self, callback):
        self.callbacks.discard(callback)

    def fire(self, **attrs):
        e = Event(source=self, **attrs)
        for fn in self.callbacks:
            fn(e)


if __name__ == '__main__':

    def callback1(event):
        print 'Callback #1 got fired with: ' + event.message

    def callback2(event):
        print 'Callback #2 got fired with: ' + event.message

    o = Trigger()
    o.subscribe(callback1)
    o.subscribe(callback2)
    o.fire(message='first trigger')
    o.unsubscribe(callback1)
    o.fire(message='second trigger')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals, absolute_import
 
from csv import DictReader
from collections import OrderedDict


class Modele(object):
 
    def __init__(self):
        self.hunks = []
        with open(csv) as f:
            # on parse le csv
            for data in DictReader(f, delimiter=b';', quotechar=b'"'):
                # on calcule les stats pour que ligne du csv
                support = self.supports.setdefault(data['Support'], {})
                support['nombre_de_jeux'] = support.get('nombre_de_jeux', 0) + 1
                self.total_jeux += 1
                if support.get('joueurs_max', 0) < data['Nombre de joueurs Max']:
                    support['joueurs_max'] = data['Nombre de joueurs Max']
 
    def __iter__(self):
        # goodies pour pouvoir itérer sur le modèle
        return self.supports.iteritems()


if __name__ == '__main__':

    import rungit

    filepath = rungit.git_changed()[0]
    modele = Modele(rungit.git_diff(filepath))
    modele.total_jeux

    for support, data in modele:
        print support
        print data



# -*- coding: utf-8 -*-
__version__ = '0.4.3'

__version_text__ = '''giterm version {0} -
    Copyright © 2015-2016 Tim Legrand -
    License BSD 2-Clause License -
    This is free software; see the source for copying conditions. There is NO
    warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE
    '''.format(__version__)



# -*- coding: utf-8 -*-
import curses
import threading

from panel import Panel, PanelManager
from postponer import Postponer
import rungit
import textutils


class GitermPanelManager(PanelManager):

    def __init__(self, stdscr):
        super(GitermPanelManager, self).__init__(stdscr)
        self.create_panels()

    def create_panels(self):
        """Creates a graphical-like UI:
    ┌────────┐┌────────────────────────────────┐
    │Branches││ Log history                    │
    │> master││                                │
    │> devel ││                                │
    │        ││                                │
    │Remotes ││                                │
    │> origin│└────────────────────────────────┘
    │        │┌───────────────┐┌───────────────┐
    │Tags    ││ Staged files  ││               │
    │        │└───────────────┘│ Diff of       │
    │Stashes │┌───────────────┐│ selected file │
    │        ││ Changed files ││               │
    └────────┘└───────────────┘└───────────────┘
        """
        height, width = self.stdscr.getmaxyx()
        if height < 8 or width < 40:
            raise Exception("Height and width must be at least 8x40.\
                Currently: %sx%s" % (height, width))
        # Following sizes are percentages (e.g. w_30 is 30% of screen width)
        w_20 = min(max(width // 5, 20), 25)
        w_30 = width // 3
        w_50 = width - w_30 - w_20
        h_49 = height // 2
        h_51 = height - h_49
        h_25 = h_51 // 2
        h_26 = h_51 - h_25
        h_l = height // 5
        h_br = height - 4 * h_l
        self['branches'] = StateLinePanel(
            self.stdscr,
            h_br, w_20, 0, 0, title='Branches')
        self['remotes'] = Panel(
            self.stdscr,
            h_l, w_20, h_br, 0, title='Remotes')
        self['tags'] = StateLinePanel(
            self.stdscr,
            h_l, w_20, h_br + h_l, 0, title='Tags')
        self['stashes'] = Panel(
            self.stdscr,
            h_l, w_20, h_br + 2 * h_l, 0, title='Stashes')
        self['submodules'] = Panel(
            self.stdscr,
            h_l, w_20, h_br + 3 * h_l, 0, title='Submodules')
        self['history'] = StateLinePanel(
            self.stdscr,
            h_49, w_30 + w_50, 0, w_20, title='History')
        self['stage'] = StagerUnstager(
            self, self.stdscr,
            h_25, w_30, h_49, w_20, title='Staging Area')
        self['changes'] = StagerUnstager(
            self, self.stdscr,
            h_26, w_30, h_49 + h_25, w_20, title='Changes')
        self['diff'] = Diff(
            self.stdscr,
            h_51, w_50, h_49, w_20 + w_30, title='Diff View')

        self['changes'].rungit = rungit.git_changed
        self['stage'].rungit = rungit.git_staged
        self['history'].rungit = rungit.git_history
        self['branches'].rungit = rungit.git_branches
        self['remotes'].rungit = rungit.git_remotes
        self['stashes'].rungit = rungit.git_stashes
        self['submodules'].rungit = rungit.git_submodules
        self['tags'].rungit = rungit.git_tags
        self['diff'].rungit = rungit.git_diff

        self['changes'].action = self.stage_file
        self['stage'].action = self.unstage_file

    def stage_file(self, path):
        rungit.git_stage_file(path)
        self['stage'].handle_event(None)
        self['changes'].handle_event(None)

    def unstage_file(self, path):
        if not path:
            raise Exception('path is mandatory')
        rungit.git_unstage_file(path)
        self['stage'].handle_event(None)
        self['changes'].handle_event(None)


class Diff(Panel):

    def __init__(self, *args, **kwargs):
        super(Diff, self).__init__(*args, **kwargs)
        self.default_title = self.title
        self.running = threading.Lock()

    def setup_content(self):
        if not self.data:
            return
        self.content = []
        cut_line = '-' * (self.CW // 2 - 1) + '8<'  + '-' * (self.CW // 2 - 1)
        for h in self.data:
            self.content += textutils.remove_superfluous_alineas(h)
            self.content.append(cut_line)

    def handle_event(self, filepath, staged=False):
        self.running.acquire()
        if filepath is None or type(filepath) is not str:
            self.running.release()
            return
        self.data = self.rungit(filepath, staged)
        self.setup_content()
        self.topLineNum = 0
        self.selected_line = -1
        self.hovered_line = 0
        message = ": " + filepath if type(filepath) == str else ''
        self.title = self.default_title + message
        self.display()
        self.running.release()

    def activate(self):
        self.cursor_y = 1
        super(Diff, self).activate()
        return self


class StagerUnstager(Panel):

    def __init__(self, parent, *args, **kwargs):
        super(StagerUnstager, self).__init__(*args, **kwargs)
        self.parent = parent
        self.default_title = self.title
        self.postponer = Postponer(timeout_in_seconds=0.3)

    def filename_from_linenum(self, linenum):
        if linenum < 0 or linenum >= len(self.content):
            return ''
        line = None
        try:
            line = self.content[linenum]
        except:
            return ''
        return line.split()[1]

    def move_cursor(self):
        super(StagerUnstager, self).move_cursor()
        if self.active:
            self.request_diff_in_diff_view()

    def handle_event(self, event=None):
        super(StagerUnstager, self).handle_event()
        self.request_diff_in_diff_view()

    def activate(self):
        this = super(StagerUnstager, self).activate()
        self.request_diff_in_diff_view()
        return this

    def request_diff_in_diff_view(self, even_not_active=False):
        if not self.active and not even_not_active:
            return
        self.hovered_line = self.cursor_y + self.topLineNum - self.CT
        if self.hovered_line < 0 or self.hovered_line >= len(self.content):
            return
        filepath = self.filename_from_linenum(self.hovered_line)
        if self.content and filepath:
            self.postponer.set(
                action=self.parent['diff'].handle_event,
                args=[filepath, (self.title == 'Staging Area')])

    def select(self):
        if self.selected_line == self.hovered_line:
            self.selected_line = -1
        else:
            self.selected_line = self.hovered_line
        if self.selected_line != -1:
            self.selected_file = self.filename_from_linenum(self.selected_line)
            self.cursor_y = max(self.cursor_y - 1, self.CT)
            self.action(self.selected_file)
            self.unselect()
        self.display()


class StateLinePanel(Panel):

    def handle_event(self, event):
        self.content = self.rungit()
        for i, line in enumerate(self.content):
            if line.startswith('*'):
                self.decorations[i] = curses.A_BOLD
                self.content[i] = line[1:]
        self.display()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import curses
import argparse
import os

import watch
import rungit
import cursutils

from gui import GitermPanelManager
from _version import __version_text__
from exception import *


def keyloop(stdscr):
    panels = GitermPanelManager(stdscr)
    active = panels['history'].activate()
    panels.display()

    w = watch.Watcher()
    for name, panel in panels.iteritems():
        w.event_handler.subscribe(panel.handle_event)

    # Initialize contents
    w.event_handler.fire()
    panels['changes'].request_diff_in_diff_view(even_not_active=True)

    w.start()

    while True:
        c = stdscr.getch()
        if 0 < c < 256:
            c = chr(c)
            if c in ' \n':  # 'SPACE BAR' or 'ENTER' hit
                active.select()
            elif c in 'Qq':
                break
            elif c == '\t':
                active = panels.toggle()
            elif c == 'h':
                active.deactivate()
                active = panels['history'].activate()
            elif c == 'c':
                active.deactivate()
                active = panels['changes'].activate()
            elif c == 's':
                active.deactivate()
                active = panels['stage'].activate()
            elif c == 'd':
                active.deactivate()
                active = panels['diff'].activate()
            elif c == 'b':
                active.deactivate()
                active = panels['branches'].activate()
            elif c == 'r':
                active.deactivate()
                active = panels['remotes'].activate()
            elif c == 't':
                active.deactivate()
                active = panels['tags'].activate()
        elif c == curses.KEY_BTAB:
            active = panels.toggle(reverse=True)
        elif c == curses.KEY_UP:
            active.move_up()
        elif c == curses.KEY_LEFT:
            active.move_left()
        elif c == curses.KEY_DOWN:
            active.move_down()
        elif c == curses.KEY_RIGHT:
            active.move_right()
        elif c == curses.KEY_PPAGE:
            active.move_prev_page()
        elif c == curses.KEY_NPAGE:
            active.move_next_page()
        elif c == curses.KEY_RESIZE:
            pass  # TODO: handle terminal resize properly (downsize and upsize)
        else:
            pass

    w.stop()


def main(stdscr):
    cursutils.init(stdscr)
    current_dir = os.getcwd()
    try:
        git_root_dir = rungit.git_root_path()
        os.chdir(git_root_dir)
        keyloop(stdscr)
    except Exception as e:
        cursutils.finalize(stdscr)
        t = type(e)
        if t == NotAGitRepositoryException or t == GitNotFoundException:
            print e
        else:
            raise
    finally:
        os.chdir(current_dir)


def _main():
    parser = argparse.ArgumentParser(
        description='''A terminal-based GUI client for Git.''')
    parser.add_argument(
        '-v', '--version', action='version', version=__version_text__)
    parser.parse_args()
    curses.wrapper(main)


if __name__ == '__main__':
    _main()



# -*- coding: utf-8 -*-
import commands
import os

import textutils

from exception import *


def run(cmd):
    code, output = commands.getstatusoutput(cmd)
    return code, [x for x in output.split('\n') if x]


def git_root_path():
    error, output = run('git rev-parse --show-toplevel')
    if error:
        if error == 32512:
            raise GitNotFoundException(
                'Git not found. Please install Git first.')
        else:
            raise NotAGitRepositoryException(
                'Please cd in a Git repository first.')
    return output[0]


def git_status(staged=False):
    error, output = run('git status -s --porcelain')
    data = []
    for line in output:
        path = line.split()[1]
        full_code = line[0:2]
        if staged:
            code = full_code[0]
        else:
            code = full_code[1]
        if code in 'CRADUM':
            data.append(line)
        elif code in ' ':
            pass
        elif code in '?!':
            if not staged:
                if os.path.isdir(path):
                    error, paths = run('find %s -type f' % path)
                    data += [full_code + ' ' + x for x in paths]
                else:
                    data.append(line)
        else:
            data.append('X' + line)
    return data


def git_changed():
    return git_status(staged=False)


def git_staged():
    return git_status(staged=True)


def git_history():
    error, data = run('git log --abbrev-commit --decorate --date=iso')
    commits = textutils.blocks(data, lambda x: x and x[:6] == 'commit')
    output = []
    for commit in commits:
        text = []
        for line in commit:
            if line.startswith('commit'):
                sha1 = line.split()[1]
                main = line.split('(', 1)
                branches = '(' + main[1] if len(main) == 2 else ''
            elif line.startswith('Author'):
                author = line.split(' ', 1)[1].lstrip()
            elif line.startswith('Date'):
                date = line.split(' ', 1)[1].lstrip()
            elif line.startswith('Merge'):
                pass
            else:
                text.append(line.lstrip())
        message = ' | '.join([x for x in text if x])
        history_line = ' '.join([branches, message, author, date, sha1])
        output.append(history_line.lstrip())
    output[0] = '*' + output[0]
    return output


def git_branches():
    error, data = run('git branch')
    for i, line in enumerate(data):
        data[i] = line[2:] if line[0] != '*' else '*' + line[2:]
    return data


def git_stashes():
    error, data = run('git stash list')
    for i, line in enumerate(data):
        data[i] = line[14:]
    return data


def git_remotes():
    error, data = run('git remote show')
    return data


def git_submodules():
    error, data = run('git submodule status')
    for i, line in enumerate(data):
        data[i] = line.split()[1]
    return data


def git_tags():
    error, data = run('git tag')
    # If Git >= 2.3.3:
    # git log --date-order --tags --simplify-by-decoration --pretty=format:"%d"
    return data


def git_diff(path, cached=False):
    if not path or type(path) is not str:
        raise Exception('Path not supported: ' + str(path))

    opt = '--cached' if cached else ''
    cmd = 'git diff {} -- {}'.format(opt, path)
    error, data = run(cmd)
    if not data:
        cmd = 'git diff -- /dev/null %s' % path
        error, data = run(cmd)
        if data:
            data = data[5:]
            error = 0
    else:
        data = data[4:]
    if error:
        raise Exception('Error executing "' + cmd + '" (error = ' + str(error))
    hunks = list(textutils.blocks(data, lambda x: x and x.startswith('@@')))
    return hunks


def run_simple_command(command, path):
    if not path:
        raise ArgumentException('File name missing.')
    command = 'git %s -- %s' % (command, path)
    code, output = commands.getstatusoutput(command)
    if code != 0:
        raise Exception('Error %s while running "%s"' % (code, command))


def git_stage_file(path):
    run_simple_command('add', path)


def git_unstage_file(path):
    run_simple_command('reset', path)


if __name__ == '__main__':
    print git_root_path()
    print git_changed()
    print git_staged()
    print git_history()
    print git_branches()
    print git_stashes()
    print git_remotes()
    print git_submodules()
    print git_tags()
    print git_diff(path='src/giterm/rungit.py')



# -*- coding: utf-8 -*-
import locale


locale.setlocale(locale.LC_ALL, '')
code = locale.getpreferredencoding()


def shorten(string, size):
    printable = string.decode(code)
    # Is that really efficient? Shouldn't we store the raw data
    # in self.content instead?
    if len(printable) > size:
        printable = printable[:size - 3] + '...'
    return printable.encode(code), len(string)


def blocks(iterable, start_pattern):
    block = []
    for line in iterable:
        if start_pattern(line):
            if block:
                yield block
                block = []
        block.append(line)
    if block:
        yield block


def remove_superfluous_alineas(hunk):
    min_alinea = 1000
    for i, line in enumerate(hunk[1:]):
        line = tabs_to_spaces(line, num_spaces=2)
        hunk[i + 1] = line
        min_alinea = get_new_minimum_alinea(line, min_alinea, 1)
    left_aligned_hunk = lstrip_hunk(hunk, min_alinea)
    return left_aligned_hunk


def tabs_to_spaces(text, num_spaces=2):
    return (num_spaces * ' ').join(text.split('\t'))


def get_new_minimum_alinea(line, previous_alinea, num_ignored_chars):
    line = line[1:]  # First char is diff-specific ('+' or '-' chars)
    lstripped_length = len(line.lstrip())
    if lstripped_length != 0:
        curr_alinea = len(line) - lstripped_length
        min_alinea = min(previous_alinea, curr_alinea)
        return min_alinea
    return previous_alinea


def lstrip_hunk(text, offset):
    '''Left-shifts of 'offset' chars every line of a given 'text',
    but saves the first character (typically '+' or '-' in diff output).
    '''
    result = []
    result.append(text[0])
    for line in text[1:]:
        result.append(line[0] + line[offset + 1:])
    return result



# -*- coding: utf-8 -*-
from threading import Timer


class Postponer(object):
    '''Fires a callback on timeout, unless timer is reset by new instructions.
    '''
    def __init__(self, timeout_in_seconds=None, action=None, *args, **kwargs):
        self.timeout = timeout_in_seconds if timeout_in_seconds else 1.0
        self.handler = action if action else self.defaultHandler
        self.timer = Timer(self.timeout, self.handler)
        # self.set(self.timeout, self.handler, *args, **kwargs)

    def set(self, timeout_in_seconds=None, action=None, *args, **kwargs):
        '''Fires a callback with its arguments on timeout.
        If called again before timeout, registers new instructions and
        resets timer.
        '''
        self.timer.cancel()

        if action:
            self.handler = action
        if timeout_in_seconds:
            self.timeout = timeout_in_seconds

        self.timer = Timer(self.timeout, self.handler, *args, **kwargs)
        self.timer.start()

    def reset(self, timeout_in_seconds=None, action=None, *args, **kwargs):
        self.set(timeout_in_seconds, action)

    def cancel(self):
        self.timer.cancel()

    def defaultHandler(self):
        raise Exception('Timeout reached!')


if __name__ == '__main__':

    import time

    def cancel_with_logging(self):
        self.timer.cancel()
        print 'Timer reset.'

    Postponer.cancel = cancel_with_logging

    def print_args(args, kwargs):
        if args or kwargs:
            arguments = [x for x in args]
            arguments += [k + '=' + x for k, x in kwargs.iteritems()]
        else:
            arguments = 'no arguments'
        return arguments

    def please_work(*args, **kwargs):
        print 'please_work() executed with: %s' % print_args(args, kwargs)

    def no_do_that_instead(*args, **kwargs):
        print 'no_that_instead() executed with: %s' % print_args(args, kwargs)

    # Instanciation does not start the timer
    p = Postponer(timeout_in_seconds=1.0)
    p.set()  # start timer for one second
    time.sleep(0.5)
    p.set(action=please_work)  # new callback, timer reset
    time.sleep(0.5)
    # Let's input new parameters, the timer resets
    p.set(
        timeout_in_seconds=0.6,
        args=['arg#1', 'arg#2'],
        kwargs={'file': 'this'})
    time.sleep(1.5)
    # Let's input new parameters, the timer resets
    p.set(timeout_in_seconds=0.6, action=no_do_that_instead)
    time.sleep(0.5)
    # Timer resets again
    p.set(timeout_in_seconds=0.6)
    time.sleep(0.7)
    # Ang again
    p.set(timeout_in_seconds=0.6, args=['no this file instead'])
    time.sleep(0.5)



# -*- coding: utf-8 -*-
import sys

from watchdog.observers import Observer
from watchdog.events import FileSystemEventHandler

from observer import Trigger


# Prevent tracking '.goutputstream-*' files (known Ubuntu bug)
GIT_BLACK_LIST = ['.git/index.lock', '.goutputstream']
GIT_WHITE_LIST = ['.gitignore', '.gitconfig', '.gitmodules']


class FileChangedHandler(FileSystemEventHandler, Trigger):
    def on_any_event(self, event):
        path = event.src_path
        if path.startswith('./'):
            path = path[2:]
        if path == '.git':
            # Not interested by the '.git' folder itself, only its contents
            return
        for forbidden in GIT_BLACK_LIST:
            if path.startswith(forbidden):
                if path not in GIT_WHITE_LIST:
                    return
                else:
                    break
        message = event.event_type + ': ' + path
        message = message[0].upper() + message[1:]
        self.action(message)

    def action(self, msg):
        if self.callbacks:
            self.fire(content=msg)
        else:
            print msg


class Watcher(object):
    def __init__(self, path='.'):
        super(Watcher, self).__init__()
        self.path = path
        self.event_handler = FileChangedHandler()
        self.observer = Observer(timeout=0.4)
        self.observer.schedule(self.event_handler, self.path, recursive=True)

    def start(self):
        self.observer.start()

    def stop(self):
        self.observer.stop()
        self.observer.join()


if __name__ == '__main__':
    import time
    watchdir = sys.argv[1] if len(sys.argv) > 1 else '.'
    w = Watcher()
    w.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        w.stop()



#!/usr/bin/env python
from setuptools import setup

exec(open('src/giterm/_version.py').read())

with open("requirements.txt") as f:
    required = [l for l in f.read().splitlines() if not l.startswith("#")]

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='giterm',
    version=__version__,
    description='A terminal-based GUI client for Git',
    long_description=long_description,
    keywords='git, client, terminal, console',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console :: Curses',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Natural Language :: English',
        'Topic :: Software Development :: Version Control',
        ],
    author='Tim Legrand',
    author_email='timlegrand.perso+dev@gmail.com',
    url='https://github.com/timlegrand/giterm',
    download_url='https://github.com/timlegrand/giterm',
    license='BSD 2-Clause',
    packages=['giterm'],
    package_dir={'giterm': 'src/giterm'},
    install_requires=required,
    entry_points={'console_scripts': ['giterm = giterm.giterm:_main']},
    )



# -*- coding: utf-8 -*-
import curses

from itertools import cycle
from collections import OrderedDict

import textutils


class PanelManager(OrderedDict):

    def __init__(self, stdscr):
        super(PanelManager, self).__init__()
        self.stdscr = stdscr

    def toggle(self, reverse=False):
        if reverse:
            items = self.items()
            items.reverse()
            reverse = OrderedDict(items)
            it = cycle(reverse.iteritems())
        else:
            it = cycle(self.iteritems())
        for k, panel in it:
            if panel.active:
                panel.deactivate()
                return next(it)[1].activate()

    def display(self):
        curses.curs_set(0)
        active = None
        for k, panel in self.iteritems():
            panel.display()
            if panel.active:
                active = panel
        if active:
            self.stdscr.move(*active.getcontentyx())


class Panel(object):
    """Encapsulates a (sub-)window.
           < W >
    0,0▒▒▒▒▒ T ▒▒▒▒▒▒▒▒  <── border
    ▒       CT        ▒  ┐
  ^ ▒CL             CR▒  │ content goes from
  H L                 R  │ (CT, CL) to
  v ▒cursor(y,x)▒     ▒  │ (CB, CR)
    ▒       CB        ▒  ┘
    ▒▒▒▒▒▒▒▒ B ▒▒▒▒▒H,W
    """

    def __init__(self, stdscr, h, w, y, x, title=''):
        self.content = []
        self.data = []
        self.decorations = {}
        self.window = stdscr.derwin(h, w, y, x)
        self.title = title
        self.H, self.W = self.window.getmaxyx()
        self.T, self.L, self.B, self.R = 0, 0, h - 1, w - 1  # Window-relative
        self.CT, self.CL, self.CB, self.CR, self.CH, self.CW\
            = self.T + 1, self.L + 1, self.B - 1, self.R - 1, h - 2, w - 2
        self.cursor_y, self.cursor_x = self.CT, self.CL  # Window-relative
        self.middle = (self.H // 2, self.W // 2)
        self.active = False
        self.topLineNum = 0
        self.selected_line = -1  # Content-relative line number [0..N-1]
        self.hovered_line = 0  # Content-relative line number [0..N-1]

    def display(self):
        self.window.erase()
        self.draw_borders()
        self.draw_content()
        self.check_cursor_position()
        self.draw_hover()
        self.draw_selected()
        self.window.refresh()

    def draw_content(self):
        top = self.topLineNum
        bottom = self.topLineNum + self.CH
        for i, line in enumerate(self.content[top:bottom]):
            y = i + self.CT
            self.add_content_line(y, line)
            index = y + self.topLineNum - self.CT
            if index in self.decorations:
                self.window.chgat(y, self.CL, self.CR, self.decorations[index])
            # TODO: need to handle case of last line fulfilled when
            # scrolling disabled

    def check_cursor_position(self):
        self.allowed_cursor_range_start = min(self.CT, len(self.content))
        self.allowed_cursor_range_end = max(
            self.allowed_cursor_range_start,
            len(self.content) - 1 - self.topLineNum + self.CT)
        if self.cursor_y < self.allowed_cursor_range_start:
            self.cursor_y = self.allowed_cursor_range_start
        elif self.cursor_y > self.allowed_cursor_range_end:
            self.cursor_y = self.allowed_cursor_range_end

    def draw_hover(self):
        self.hovered_line = self.cursor_y + self.topLineNum - self.CT
        y = self.cursor_y
        if (self.active and y >= self.CT and y <= self.CB and self.content):
            index = y + self.topLineNum - self.CT
            mode = curses.A_NORMAL
            if index in self.decorations:
                mode = self.decorations[index]
            self.window.chgat(y, self.CL, self.CR, mode | curses.A_REVERSE)

    def draw_selected(self):
        if self.selected_line != -1:
            y = self.selected_line + self.CT - self.topLineNum
            char = self.window.inch(y, self.CL)
            attr = char & curses.A_ATTRIBUTES
            self.window.chgat(y, self.CL, self.CR, attr | curses.A_BOLD)

    def add_content_line(self, line_num, content):
        short, num_raw_bytes = textutils.shorten(content, self.CW)
        self.window.addnstr(line_num, self.CL, short, num_raw_bytes)

    def draw_borders(self):
        self.window.box()
        top, left, title, n = self.T, self.L + 2, self.title, self.W - 4
        if title:
            self.window.addnstr(top, left, ' ' + title + ' ', n)
            if self.active:
                self.window.addnstr(top, left, '[' + title + ']', n + 2)
                self.window.chgat(top, left, len(title) + 2, curses.A_BOLD)
        sidebar_pos = int(self.topLineNum * self.CH /
                          (max(len(self.content) - self.CH, 1)))
        if len(self.content) > self.CH:
            if sidebar_pos < self.CT:
                sidebar_pos = sidebar_pos + self.CT
            if sidebar_pos > self.CB:
                sidebar_pos = self.CB
            self.window.addnstr(sidebar_pos, self.R, 'o', 1)

    def setup_content(self):
        if self.data:
            self.content = self.data

    # Callback function for remote observers
    def handle_event(self, event=None):
        self.data = self.rungit()
        self.setup_content()
        self.display()

    def select(self):
        if self.hovered_line == self.selected_line:
            self.selected_line = -1
        else:
            self.selected_line = self.hovered_line
        self.display()

    def unselect(self):
        self.selected_line = -1
        self.display()

    def activate(self):
        if self.active:
            return
        self.active = True
        self.display()
        return self

    def deactivate(self, force=False):
        if not self.active and not force:
            return
        self.active = False
        self.display()

    def getcontentyx(self):
        y, x = self.window.getbegyx()
        return y + self.CT, x + self.CL

    def text_center(self, row, col, string):
        self.window.addstr(row, col - len(string) / 2, string)

    def text(self, y, x, string):
        self.window.addstr(y, x, string)

    def text_right_align(self, y, x, string):
        self.window.addstr(y, x - len(string) + 1, string)

    def text_force_right_align(self, y, x, string):
        '''Forces right-aligned text to be printed
        until the last char position of the panel
        even with scrolling disabled'''
        try:
            self.window.addstr(y, x - len(string) + 1, string)
        except curses.error:
            pass

    def move_left(self):
        if self.cursor_x > self.CL:
            self.cursor_x -= 1
            self.move_cursor()

    def move_right(self):
        if self.cursor_x < self.CR:
            self.cursor_x += 1
            self.move_cursor()

    def move_up(self):
        if self.cursor_y > self.CT:
            self.cursor_y -= 1
        elif self.topLineNum > 0:
            self.topLineNum -= 1
        else:
            return
        self.move_cursor()

    def move_down(self):
        if self.cursor_y < self.CB and self.cursor_y < len(self.content):
            self.cursor_y += 1
        elif self.topLineNum + self.CH < len(self.content):
            self.topLineNum += 1
        else:
            return
        self.move_cursor()

    def move_prev_page(self):
        if self.topLineNum < self.CH:
            self.topLineNum = 0
        else:
            self.topLineNum -= self.CH
        self.hovered_line = self.topLineNum
        self.move_cursor()

    def move_next_page(self):
        if self.topLineNum + self.CH < len(self.content):
            self.topLineNum += self.CH
        else:
            return
        self.hovered_line = self.topLineNum
        self.move_cursor()

    def move_cursor(self):
        self.hovered_line = self.topLineNum + self.cursor_y - self.CT
        self.window.move(self.cursor_y, self.cursor_x)
        self.display()

    def debug(self, refresh=True):
        self.window.box(curses.ACS_CKBOARD, curses.ACS_CKBOARD)
        active = '  Active  ' if self.active else ' Inactive '
        self.text_center(0, self.W // 2, active)
        self.text_center(self.middle[0], self.middle[1], str(self.middle))
        size = '[' + str(self.H) + ' x ' + str(self.W) + ']'
        self.text_center(self.middle[0] + 1, self.middle[1], size)
        TL = str((self.T, self.L))
        self.window.addstr(self.T, self.L, TL)
        TR = str((self.T, self.R))
        self.text_force_right_align(self.T, self.R, TR)
        BL = str((self.B, self.L))
        self.window.addstr(self.B, self.L, BL)
        BR = str((self.B, self.R))
        self.text_force_right_align(self.B, self.R, BR)
        if refresh:
            self.window.refresh()
